@@ -23,6 +23,7 @@
 struct _SoupFilterInputStreamPrivate {
 	GByteArray *buf;
 	gboolean need_more;
+	gboolean in_read_until;
 };
 
 static void soup_filter_input_stream_pollable_init (GPollableInputStreamInterface *pollable_interface, gpointer interface_data);
@@ -79,8 +80,10 @@ soup_filter_input_stream_read_fn (GInputStream  *stream,
 {
 	SoupFilterInputStream *fstream = SOUP_FILTER_INPUT_STREAM (stream);
 
-	fstream->priv->need_more = FALSE;
-	if (fstream->priv->buf) {
+	if (!fstream->priv->in_read_until)
+		fstream->priv->need_more = FALSE;
+
+	if (fstream->priv->buf && !fstream->priv->in_read_until) {
 		return read_from_buf (fstream, buffer, count);
 	} else {
 		return g_pollable_stream_read (G_FILTER_INPUT_STREAM (fstream)->base_stream,
@@ -108,8 +111,10 @@ soup_filter_input_stream_read_nonblocking (GPollableInputStream  *stream,
 {
 	SoupFilterInputStream *fstream = SOUP_FILTER_INPUT_STREAM (stream);
 
-	fstream->priv->need_more = FALSE;
-	if (fstream->priv->buf) {
+	if (!fstream->priv->in_read_until)
+		fstream->priv->need_more = FALSE;
+
+	if (fstream->priv->buf && !fstream->priv->in_read_until) {
 		return read_from_buf (fstream, buffer, count);
 	} else {
 		return g_pollable_stream_read (G_FILTER_INPUT_STREAM (fstream)->base_stream,
@@ -196,7 +201,7 @@ soup_filter_input_stream_read_until (SoupFilterInputStream  *fstream,
 				     GCancellable           *cancellable,
 				     GError                **error)
 {
-	gssize nread;
+	gssize nread, read_length;
 	guint8 *p, *buf, *end;
 	gboolean eof = FALSE;
 	GError *my_error = NULL;
@@ -217,10 +222,12 @@ soup_filter_input_stream_read_until (SoupFilterInputStream  *fstream,
 		g_byte_array_set_size (fstream->priv->buf, length);
 		buf = fstream->priv->buf->data;
 
-		nread = g_pollable_stream_read (G_FILTER_INPUT_STREAM (fstream)->base_stream,
+		fstream->priv->in_read_until = TRUE;
+		nread = g_pollable_stream_read (G_INPUT_STREAM (fstream),
 						buf + prev_len, length - prev_len,
 						blocking,
 						cancellable, &my_error);
+		fstream->priv->in_read_until = FALSE;
 		if (nread <= 0) {
 			if (prev_len)
 				fstream->priv->buf->len = prev_len;
@@ -247,10 +254,11 @@ soup_filter_input_stream_read_until (SoupFilterInputStream  *fstream,
 	} else
 		buf = fstream->priv->buf->data;
 
-	/* Scan for the boundary */
-	end = buf + fstream->priv->buf->len;
-	if (!eof)
-		end -= boundary_length;
+	/* Scan for the boundary within the range we can possibly return. */
+	if (include_boundary)
+		end = buf + MIN (fstream->priv->buf->len, length) - boundary_length;
+	else
+		end = buf + MIN (fstream->priv->buf->len - boundary_length, length);
 	for (p = buf; p <= end; p++) {
 		if (*p == *(guint8*)boundary &&
 		    !memcmp (p, boundary, boundary_length)) {
@@ -264,10 +272,9 @@ soup_filter_input_stream_read_until (SoupFilterInputStream  *fstream,
 	if (!*got_boundary && fstream->priv->buf->len < length && !eof)
 		goto fill_buffer;
 
-	/* Return everything up to 'p' (which is either just after the boundary if
-	 * include_boundary is TRUE, just before the boundary if include_boundary is
-	 * FALSE, @boundary_len - 1 bytes before the end of the buffer, or end-of-
-	 * file).
-	 */
-	return read_from_buf (fstream, buffer, p - buf);
+	if (eof && !*got_boundary)
+		read_length = MIN (fstream->priv->buf->len, length);
+	else
+		read_length = p - buf;
+	return read_from_buf (fstream, buffer, read_length);
 }
